@@ -18,8 +18,12 @@
 #include <limits>
 #include <stdexcept>
 #include <string>
+#if __cplusplus >= 201703L
+#include <string_view>
+#endif
 #include <type_traits>
 #include <utility>
+#include <vector>
 
 #include <mp++/concepts.hpp>
 #include <mp++/config.hpp>
@@ -73,13 +77,22 @@ inline void fix_den_sign(rational<SSize> &q)
         q._get_den().neg();
     }
 }
+
+// Detect rational.
+template <typename T>
+struct is_rational : std::false_type {
+};
+
+template <std::size_t SSize>
+struct is_rational<rational<SSize>> : std::true_type {
+};
 }
 
 /// Multiprecision rational class.
 /**
  * \rststar
  * This class represents arbitrary-precision rationals. Internally, the class stores a pair of
- * :cpp:class:`integers <mppp::integer>` with static size ``SSize`` as the numerator and denonimator.
+ * :cpp:class:`integers <mppp::integer>` with static size ``SSize`` as the numerator and denominator.
  * Rational numbers are represented in the usual canonical form:
  *
  * * numerator and denominator are coprime,
@@ -150,6 +163,18 @@ inline void fix_den_sign(rational<SSize> &q)
  *    auto res5 = q1 / 2.f; // res5 is a float
  *    auto res6 = 12. / q1; // res6 is a double
  *
+ * The relational operators, ``==``, ``!=``, ``<``, ``>``, ``<=`` and ``>=`` will promote the arguments to a common type
+ * before comparing them. The promotion rules are the same as in the arithmetic operators (that is, both arguments are
+ * promoted to :cpp:class:`~mppp::rational` if no floating-point types are involved, otherwise they are promoted to the
+ * type of the floating-point argument).
+ *
+ * The :cpp:class:`~mppp::rational` class allows to access and manipulate directly the numerator and denominator
+ * via the :cpp:func:`~mppp::rational::get_num()`, :cpp:func:`~mppp::rational::get_den()`,
+ * :cpp:func:`~mppp::rational::_get_num()` and :cpp:func:`~mppp::rational::_get_den()` methods, so that it is possible
+ * to use :cpp:class:`~mppp::integer` functions directly on numerator and denominator. The mutable getters' names
+ * :cpp:func:`~mppp::rational::_get_num()` and :cpp:func:`~mppp::rational::_get_den()` are prefixed with an underscore
+ * ``_`` to highlight their potentially dangerous nature: it is the user's responsibility to ensure that the canonical
+ * form of the rational is preserved after altering the numerator and/or the denominator via the mutable getters.
  * \endrststar
  */
 // NOTEs:
@@ -248,8 +273,8 @@ private:
         }
         MPPP_MAYBE_TLS mpq_raii q;
         ::mpq_set_d(&q.m_mpq, static_cast<double>(x));
-        m_num.dispatch_mpz_ctor(mpq_numref(&q.m_mpq));
-        m_den.dispatch_mpz_ctor(mpq_denref(&q.m_mpq));
+        m_num = mpq_numref(&q.m_mpq);
+        m_den = mpq_denref(&q.m_mpq);
     }
 #if defined(MPPP_WITH_MPFR)
     void dispatch_generic_construction(const long double &x)
@@ -268,8 +293,8 @@ private:
         ::mpfr_set_ld(&mpfr.m_mpfr, x, MPFR_RNDN);
         ::mpfr_get_f(&mpf.m_mpf, &mpfr.m_mpfr, MPFR_RNDN);
         ::mpq_set_f(&mpq.m_mpq, &mpf.m_mpf);
-        m_num.dispatch_mpz_ctor(mpq_numref(&mpq.m_mpq));
-        m_den.dispatch_mpz_ctor(mpq_denref(&mpq.m_mpq));
+        m_num = mpq_numref(&mpq.m_mpq);
+        m_den = mpq_denref(&mpq.m_mpq);
     }
 #endif
 
@@ -327,23 +352,10 @@ public:
         }
         canonicalise();
     }
-    /// Constructor from C string.
-    /**
-     * \rststar
-     * This constructor will initialize ``this`` from the null-terminated string ``s``, which must represent
-     * a rational value in base ``base``. The expected format is either a numerator-denominator pair separated
-     * by the division operator ``/``, or just a numerator (in which case the denominator will be set to one).
-     * The format for numerator and denominator is described in the documentation of the constructor from string
-     * of :cpp:class:`~mppp::integer`.
-     * \endrststar
-     *
-     * @param s the input string.
-     * @param base the base used in the string representation.
-     *
-     * @throws zero_division_error if the denominator is zero.
-     * @throws unspecified any exception thrown by the string constructor of mppp::integer.
-     */
-    explicit rational(const char *s, int base = 10)
+
+private:
+    // Implementation of the constructor from C string. Requires a def-cted object.
+    void dispatch_c_string_ctor(const char *s, int base)
     {
         MPPP_MAYBE_TLS std::string tmp_str;
         auto ptr = s;
@@ -363,6 +375,28 @@ public:
             canonicalise();
         }
     }
+
+public:
+    /// Constructor from C string.
+    /**
+     * \rststar
+     * This constructor will initialize ``this`` from the null-terminated string ``s``, which must represent
+     * a rational value in base ``base``. The expected format is either a numerator-denominator pair separated
+     * by the division operator ``/``, or just a numerator (in which case the denominator will be set to one).
+     * The format for numerator and denominator is described in the documentation of the constructor from string
+     * of :cpp:class:`~mppp::integer`.
+     * \endrststar
+     *
+     * @param s the input string.
+     * @param base the base used in the string representation.
+     *
+     * @throws zero_division_error if the denominator is zero.
+     * @throws unspecified any exception thrown by the string constructor of mppp::integer.
+     */
+    explicit rational(const char *s, int base = 10)
+    {
+        dispatch_c_string_ctor(s, base);
+    }
     /// Constructor from C++ string (equivalent to the constructor from C string).
     /**
      * @param s the input string.
@@ -373,6 +407,69 @@ public:
     explicit rational(const std::string &s, int base = 10) : rational(s.c_str(), base)
     {
     }
+    /// Constructor from range of characters.
+    /**
+     * This constructor will initialise \p this from the content of the input half-open range,
+     * which is interpreted as the string representation of a rational in base \p base.
+     *
+     * Internally, the constructor will copy the content of the range to a local buffer, add a
+     * string terminator, and invoke the constructor from C string.
+     *
+     * @param begin the begin of the input range.
+     * @param end the end of the input range.
+     * @param base the base used in the string representation.
+     *
+     * @throws unspecified any exception thrown by the constructor from C string.
+     */
+    explicit rational(const char *begin, const char *end, int base = 10)
+    {
+        // Copy the range into a local buffer.
+        MPPP_MAYBE_TLS std::vector<char> buffer;
+        buffer.assign(begin, end);
+        buffer.emplace_back('\0');
+        dispatch_c_string_ctor(buffer.data(), base);
+    }
+#if __cplusplus >= 201703L
+    /// Constructor from string view.
+    /**
+     * This constructor will initialise \p this from the content of the input string view,
+     * which is interpreted as the string representation of a rational in base \p base.
+     *
+     * Internally, the constructor will invoke the constructor from a range of characters.
+     *
+     * \rststar
+     * .. note::
+     *
+     *   This constructor is available only if at least C++17 is being used.
+     * \endrststar
+     *
+     * @param s the \p std::string view that will be used for construction.
+     * @param base the base used in the string representation.
+     *
+     * @throws unspecified any exception thrown by the constructor from C string.
+     */
+    explicit rational(const std::string_view &s, int base = 10) : rational(s.data(), s.data() + s.size(), base)
+    {
+    }
+#endif
+    /// Constructor from \p mpz_t.
+    /**
+     * This constructor will initialise the numerator of \p this with the input GMP integer \p n,
+     * and the denominator to 1.
+     *
+     * \rststar
+     * .. warning::
+     *
+     *    It is the user's responsibility to ensure that ``n`` has been correctly initialized. Calling this constructor
+     *    with an uninitialized ``n`` results in undefined behaviour.
+     * \endrststar
+     *
+     * @param n the input GMP integer.
+     */
+    explicit rational(const ::mpz_t n) : m_num(n)
+    {
+        fast_set_den_one();
+    }
     /// Constructor from \p mpq_t.
     /**
      * This constructor will initialise the numerator and denominator of \p this with those of the GMP rational \p q.
@@ -382,16 +479,13 @@ public:
      *
      *    It is the user's responsibility to ensure that ``q`` has been correctly initialized. Calling this constructor
      *    with an uninitialized ``q`` results in undefined behaviour. Also, this constructor will **not**
-     *    canonicalise ``this``: numerator and denominator are constructed
-     *    as-is from ``q``.
+     *    canonicalise ``this``: numerator and denominator are constructed as-is from ``q``.
      * \endrststar
      *
      * @param q the input GMP rational.
      */
-    explicit rational(const ::mpq_t q)
+    explicit rational(const ::mpq_t q) : m_num(mpq_numref(q)), m_den(mpq_denref(q))
     {
-        m_num.dispatch_mpz_ctor(mpq_numref(q));
-        m_den.dispatch_mpz_ctor(mpq_denref(q));
     }
     /// Defaulted copy-assignment operator.
     /**
@@ -423,6 +517,31 @@ public:
         }
         return *this;
     }
+    /// Assignment from \p mpz_t.
+    /**
+     * \rststar
+     * This assignment operator will copy into the numerator of ``this`` the value of the GMP integer ``n``,
+     * and it will set the denominator to 1 via :cpp:func:`mppp::integer::set_one()`.
+     *
+     * .. warning::
+     *
+     *    It is the user's responsibility to ensure that ``n`` has been correctly initialized. Calling this operator
+     *    with an uninitialized ``n`` results in undefined behaviour. Also, no aliasing is allowed:
+     *    the data in ``n`` must be completely distinct from the data in ``this`` (e.g., if ``n`` is an ``mpz_view`` of
+     *    the numerator of ``this`` then it might point to internal data of ``this``, and the behaviour of this operator
+     *    will thus be undefined).
+     * \endrststar
+     *
+     * @param n the input GMP integer.
+     *
+     * @return a reference to \p this.
+     */
+    rational &operator=(const ::mpz_t n)
+    {
+        m_num = n;
+        m_den.set_one();
+        return *this;
+    }
     /// Assignment from \p mpq_t.
     /**
      * This assignment operator will copy into \p this the value of the GMP rational \p q.
@@ -432,7 +551,8 @@ public:
      *
      *    It is the user's responsibility to ensure that ``q`` has been correctly initialized. Calling this operator
      *    with an uninitialized ``q`` results in undefined behaviour. Also, this operator will **not** canonicalise
-     *    the assigned value: numerator and denominator are assigned as-is from ``q``.
+     *    the assigned value: numerator and denominator are assigned as-is from ``q``. Finally, no aliasing is allowed:
+     *    the data in ``q`` must be completely distinct from the data in ``this``.
      * \endrststar
      *
      * @param q the input GMP rational.
@@ -575,7 +695,7 @@ private:
     template <typename T, enable_if_t<std::is_same<bool, T>::value, int> = 0>
     std::pair<bool, T> dispatch_conversion() const
     {
-        return {true, m_num.m_int.m_st._mp_size != 0};
+        return std::make_pair(true, m_num.m_int.m_st._mp_size != 0);
     }
     // Conversion to integral types other than bool.
     template <typename T,
@@ -588,7 +708,7 @@ private:
     template <typename T, enable_if_t<disjunction<std::is_same<T, float>, std::is_same<T, double>>::value, int> = 0>
     std::pair<bool, T> dispatch_conversion() const
     {
-        return {true, static_cast<T>(::mpq_get_d(get_mpq_view()))};
+        return std::make_pair(true, static_cast<T>(::mpq_get_d(get_mpq_view())));
     }
 #if defined(MPPP_WITH_MPFR)
     // Conversion to long double.
@@ -601,7 +721,7 @@ private:
         MPPP_MAYBE_TLS mpf_raii mpf(static_cast<::mp_bitcnt_t>(d2));
         ::mpf_set_q(&mpf.m_mpf, get_mpq_view());
         ::mpfr_set_f(&mpfr.m_mpfr, &mpf.m_mpf, MPFR_RNDN);
-        return {true, ::mpfr_get_ld(&mpfr.m_mpfr, MPFR_RNDN)};
+        return std::make_pair(true, ::mpfr_get_ld(&mpfr.m_mpfr, MPFR_RNDN));
     }
 #endif
 
@@ -830,9 +950,15 @@ private:
     int_t m_den;
 };
 
+#if __cplusplus < 201703L
+
+// NOTE: see the explanation in integer.hpp regarding static constexpr variables in C++17.
+
 template <std::size_t SSize>
 constexpr std::size_t rational<SSize>::ssize;
 
+#endif
+
 inline namespace detail
 {
 
@@ -2317,7 +2443,9 @@ inline int cmp(const rational<SSize> &op1, const rational<SSize> &op2)
 template <std::size_t SSize>
 inline int cmp(const rational<SSize> &op1, const integer<SSize> &op2)
 {
-    return ::mpq_cmp_z(op1.get_mpq_view(), op2.get_mpz_view());
+    // NOTE: mpq_cmp_z() is a macro or a function, depending on the GMP version. Don't
+    // call it with "::".
+    return mpq_cmp_z(op1.get_mpq_view(), op2.get_mpz_view());
 }
 
 /// Comparison function for integer/rational arguments.
